/*
**********************************************************************
$SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
$HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/dcon.cpp $
$LastChangedRevision: 85 $
$Date: 2014-07-02 16:07:08 -0400 (Wed, 02 Jul 2014) $
$Author: agulati $
**********************************************************************

 PURPOSE :
     THIS IS A CONVOLUTION TYPE RECONSTRUCTION ALGORITHM FOR
     DIVERGENT BEAM PROJECTION DATA.  VARIOUS KINDS OF REGULARIZING
     FUNCTIONS ARE IMPLEMENTED.
 INPUT FILE :
     TWO (OR MORE) DATA CARDS HAVE TO BE SUPPLIED BY THE USER
     FOLLOWING EXECUTION COMMAND. THE INFORMATION ON AND AFTER
     THE THIRD CARD ARE FILTER PARAMETERS, THEREFORE THEY ARE
     FILTER DEPENDENT.
      CARD            LIST
      ====            ====
       1      INTERP,MISSP,NCONR,WCONV,METHOD
       2      NF
       3      C,ALPHA          -- IF(NF.NE.4HLINE
                                  .AND.NF.NE.4HBAND)
       3      X1,Y1,X2,Y2,...,XL,YL
                               -- IF(NF.EQ.4HLINE)
     DESCRIPTION OF INPUT PARAMETERS :
       INTERP - PARAMETER WHICH SPECIFIES INTERPOLATION METHOD,
                THE VALUE OF INTERP MUST LIE BETWEEN -1 AND 6 AND
                HAS THE FOLLOWING MEANING:
                -1 -- MODIFIED CUBIC SPLINE INTERPOLATION,
                 0 -- BAND-LIMITING (SINC) INTERPOLATION,
                POSITIVE -- INTERP-POINTS LAGRANGE INTERPOLATION.
       MISSP  - EVERY MISSP'TH PROJECTION IS TO BE USED IN
                RECONSTRUCTION.
       NCONR  - NUMBER OF RAYS USED ON EACH SIDE OF CONVOLUTION,
                FULL CONVOLUTION USED IF NCONR IS NEGATIVE.
       WCONV  - WEIGHT FOR CENTER RAY IN PRE-SMOOTHING.
       METHOD - PARAMETER WHICH DETERMINES THE METHOD IN FINDING
                THE RAY NUMBER AND WEIGHT FOR BACK PROJECTION.
                 0 -- ACCURATE BACK PROJECTION,
                 1 -- WEIGHT FOR BACKPROJECTION APPROXIMATED,
                -1 -- BOTH WIEGHT AND RAY NUMBER APPROXIMATED.
       NF     - FILTER NAME, CAN BE ONE OF LINE, HAMMING, SINC,
                EXPONENTIAL, PARABOLIC, COSINE, SHEPP, OR BAND.
       C      - RATIO OF FILTER BANDWIDTH TO NYQUIST CUTOFF FREQ.
       ALPHA  - PARAMETER USED IN HAMMING FILTER ONLY.
       (XN,YN)- COORDINATES WHICH DEFINE THE FILTER FUNCTION
                  SEE SNARK05 MANUAL FOR DETAIL
     EXAMPLES OF CONTROL CARD SEQUENCE :
       1)     4    1   -1  1.0    0
          HAMMING
                 1.0      0.54
       2)     2    3  100 0.75    1
          LINE
                 0.3       1.0       0.5       0.7
                 0.8       0.4       1.0       0.0
       3)     0    1   -1  0.9   -1
          BAND-LIMITING
 WARNING :
     IT IS ASSUMED THAT PROJECTION ANGLES ARE GIVEN IN INCREASING
     ORDER AND THE LAST ANGLE IS LESS THAN THE FIRST ANGLE PLUS
     TWOPI.
     IT IS THE RESPONSBILITY OF THE USER TO CHECK THAT USRAYS
     IS LARGE ENOUGH FOR THE INTERPOLATION METHOD TO WORK.
 REFERENCES :
     1) FAST IMAGE RECONSTRUCTION BASED ON A RADON INVERSION
        FORMULA APPROPRIATE FOR RAPIDLY COLLECTED DATA
        G. T. HERMAN AND A. NAPARSTEK
        SIAM JOURNAL ON APPLIED MATHEMATICS V.33 (1977)
     2) FILTER SELECTION FOR FAN-BEAM CONVOLUTION RECONSTRUCTION
        ALGORITHM
        T. CHANG AND G. T. HERMAN
        J. COMPUTER ASSISTED TOMOGRAPHY, TO APPEAR
 WRITTEN BY T. CHANG, MAY 1977
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*/

#include <cmath>
#include <cstdio>

#include "blkdta.h"
#include "geom.h"
#include "spctrm.h"
#include "consts.h"
#include "uiod.h"
#include "anglst.h"
#include "qintp.h"
#include "second.h"
#include "projfile.h"
#include "infile.h"

#include "dcon.h"

BOOLEAN dcon_class::Run(REAL* recon, INTEGER* list, REAL* weight, INTEGER iter)
{
  //fprintf(output,"\n xalg4");
  
  static const INTEGER band = CHAR2INT('b','a','n','d');
  
  // modified calls to getwrd() to use the 4-parameter version. Lajos, Dec 13, 2004
  // added missing options from "dcon_dconft.cpp". Lajos, Jan 13, 2005
  static const INTEGER dcon_codes[8] =
  {
    CHAR2INT('b','a','n','d'),
    CHAR2INT('l','i','n','e'),
    CHAR2INT('h','a','m','m'),
    CHAR2INT('p','a','r','a'),
    CHAR2INT('e','x','p','o'),
    CHAR2INT('s','i','n','c'),
    CHAR2INT('c','o','s','i'),
    CHAR2INT('s','h','e','p')
  };


  BOOLEAN eol;
  REAL theta,sinth,costh;

  REAL kpr;
  REAL* g = NULL;
  REAL* g1= NULL;

  REAL* q1 = NULL;
  REAL* q2 = NULL;
  REAL*  detecn = NULL;


  ////////////////////

  INTEGER interp;
  int i;
  INTEGER missp;
  INTEGER nconr;
  REAL wconv;
  INTEGER method;
  REAL wconw;
  INTEGER nf;
  INTEGER nraym;
  REAL fmdray;

  INTEGER mdraym;
  INTEGER mraym2;
  REAL halfl;

  REAL rcube;

  REAL alfa;

  REAL pcovst;

  REAL sigma;

  REAL ctime;

  REAL btime;

  REAL beg;

  REAL end;
  REAL ftime;
  REAL* ncos = NULL;

  REAL limmid;

  REAL* mdmod;
  REAL* modi = NULL;
  REAL* mdg;
  REAL* mdg1;
  REAL anginy;

  REAL angiin;

  REAL sbtrin;

  REAL factor;

  INTEGER lpj;
  REAL beta0;

  int np;
  REAL dtc;

  REAL datm;
  REAL dath;
  REAL datl;

  REAL cbeg;
  int k;
  REAL* mr1;
  REAL* ml1;
  REAL* mr;
  REAL* ml;
  REAL pr1;
  REAL pr2;
  REAL pl1;
  REAL pl2;
  REAL* mq1;
  REAL* mq2;
  int i1;
  REAL cend;
  REAL beta1;

  REAL sbeta;

  REAL cbeta;

  REAL w;

  REAL b;
  REAL cbhl;
  REAL sbhl;
  REAL cbpxs;

  REAL sbpxs;

  REAL usini;
  REAL ucosi;

  REAL bovrc;
  REAL wbkpjy;

  REAL wincy;

  REAL wincx;

  REAL angsin;
  REAL angcos;

  REAL anginc;

  REAL subtr;

  INTEGER ind;

  int iy;
  REAL usin;
  REAL ucos;
  REAL wbkpj;
  REAL ang;
  int ix;

  REAL bend;

  BOOLEAN ret_xalg4;

  ////////////////////
 
  if(GeoPar.nrays <= 1) {

    fprintf(output, "\n         ***** nrays too small - job aborted *****");

    return TRUE;
  }

  // ALL BUT THE FIRST ITERATION ARE FOR SMOOTHING ONLY

  ret_xalg4 = FALSE;
  if(iter > 1) return false;

  // CHECK DATA COLLECTING GEOMETRY

  if(!GeoPar.div) {
    fprintf(output, "\n      *****geometry error*****");
    fprintf(output, "\n\n     this algorithm is written for fan beam only \n\n");
    return TRUE;
  }

  // READ CHECK AND ECHO USER SUPPLIED INFORMATION

  interp = InFile.getint(TRUE, &eol);

  if((interp < -1) || (interp > 6)) {

    fprintf(output, "\n          ***** illegal value for interp *****");
    ret_xalg4 = TRUE;
  }

  i = GeoPar.nrays - GeoPar.snrays;
  if(((interp == -1) && (i < 2)) || ((interp > 2) && (i < (interp-2)))) {

    fprintf(output, "\n          ***** size of modified projection data too small");
    fprintf(output, "\n               for interpolation method specified *****");

    ret_xalg4 = TRUE;
  }

  missp = InFile.getint(FALSE, &eol);

  if((missp <= 0) || (missp >= GeoPar.prjnum)) {
    fprintf(output, "\n         ***** illegal value for missp *****");

    ret_xalg4 = TRUE;
  }


  nconr = InFile.getint(FALSE, &eol);
  wconv = InFile.getnum(FALSE, &eol);
  if((wconv <= 0) || (wconv > 1)) {
    fprintf(output, "\n         ***** illegal value for wconv *****");

        ret_xalg4 = TRUE;
  }


  method = InFile.getint(FALSE, &eol);
    
  if(eol) {
    fprintf(output, "\n         ***** insufficient data item - end of line encountered *****");

    return TRUE;
  }

  if(ret_xalg4) return ret_xalg4;

  if(interp > 0) {
    fprintf(output, "\n          %1i point lagrange interpolation", interp);
  }

  if(interp < 0) {
    fprintf(output, "\n          cubic spline interpolation");
  }

  if(interp == 0) {
    fprintf(output, "\n          band limiting (sinc) interpolation");
  }

  fprintf(output, "\n          one out of every %5i projection(s) used for reconstruction", missp);
    

  if((nconr < 0) || (nconr >= GeoPar.nrays)) nconr = GeoPar.nrays - 1;

  fprintf(output, "\n          %3i points used on each side of convolution", nconr);

  wconw = (REAL) 0.5 * ((REAL) 1.0 - wconv);
  fprintf(output, "\n          pre-smoothing weight on center ray = %6.4f", wconv);
  fprintf(output, "\n          pre-smoothing weight on side rays  = %6.4f", wconw);


  if(method > 0) {
    fprintf(output, "\n          weight for back-projection approximated");
  }

  if(method == 0) {
    fprintf(output, "\n          accurate back-projection");
  }

  if(method < 0) {
    fprintf(output, "\n          both weight and ray number for back-projection approximated");
  }

  nf = InFile.getwrd(TRUE, &eol, dcon_codes, 8);
  if(eol) {
    fprintf(output, "\n         ***** insufficient data item - end of line encountered *****");

    return TRUE;
  }

  // USE HERMAN, LAKSHMINARAYANAN AND NAPARSTEK'S ALGORITHM FOR
  // BAND-LIMITING FILTER

  if(nf == band) {
    fprintf(output, "\n          band-limiting filter");

    dconbl(recon, interp, missp, nconr, wconv, method);
    return ret_xalg4;
  }

  // INITIALIZE CONSTANTS

  nraym = GeoPar.nrays - 1;
  fmdray = (REAL) GeoPar.midray + 1;
  mdraym = GeoPar.midray;
  mraym2 = GeoPar.midray - 1;
  halfl = (REAL) GeoPar.midpix * GeoPar.pixsiz;
  rcube = GeoPar.radius * GeoPar.radius * GeoPar.radius;
  alfa = GeoPar.pinc / GeoPar.stod;

  // INITIALIZE CONSTANTS FOR TANGENT DETECTOR CASE
  // DETECN IS THE BASE FOR THE ARRAY WHICH STORES THE
  // ARC DETECTOR LOCATIONS IN TANGENT DETECTOR UNIT

  if(!(GeoPar.arc || (GeoPar.nrays <= 3))) {
    pcovst = alfa;
    alfa = (REAL) atan((REAL) mdraym * alfa) / (REAL) mdraym;

    
    detecn = new REAL[mraym2];

    sigma = alfa;
    for(i = 0; i < mraym2; i++) {
      detecn[i] = (REAL)( (REAL) sin(sigma) / (REAL) cos(sigma)) / pcovst;
      sigma += alfa;
    }
  }

  // INITIALIZE TIMERS

  ctime = 0.;
  btime = 0.;
  second(&beg);

  // ALLOCATE WORK AREAS FOR CONVOLUTING FUNCTION
  // Q1 IS THE INDEX FOR THE FIRST ELEMENT IN Q1(M)
  // Q2 IS THE INDEX FOR THE FIRST ELEMENT IN Q2(M)

  q1 = new REAL[GeoPar.nrays+1];        //(JD 11/06)
  q2 = new REAL[GeoPar.nrays+1];        //(JD 11/06)

  // CONVOLUTING FUNCTION

  if(dconft(q1, q2, nconr+1, alfa, nf)) { 
    

  if(detecn != NULL) delete[] detecn ;
  if(q1 != NULL) delete[] q1 ;
  if(q2 != NULL) delete[] q2 ;
  if(ncos != NULL) delete[] ncos ;
  if(g != NULL) delete[] g ;
  if(g1 != NULL) delete[] g1;
  if(modi != NULL) delete[] modi ;           //wei 3/2005
  
    return TRUE;
  }
  second(&end);
  ftime = end - beg;

  // WEIGHT ARRAY FOR CONVOLUTION
  // NCOS IS THE BASE FOR CONVOLUTION WEIGHT ARRAY

  ncos = new REAL[mdraym];

  sigma = alfa;

  for(i = 0; i < mdraym; i++) {
    ncos[i] = (REAL) cos(sigma);
    sigma += alfa;
  }

  // ALLOCATE AREAS FOR PROJECTION DATA
  // G IS THE BASE FOR PROJECTION DATA FOR ONE PROJECTION
  // G1 IS THE BASE FOR WEIGHTED PROJECTION DATA
  // MODI IS THE BASE FOR MODIFIED PROJECTION DATA

  g = new REAL[GeoPar.nrays];
  g1 = new REAL[GeoPar.nrays];

  modi = new REAL[GeoPar.nrays];

  limmid = (REAL) MIN0(nconr, mdraym);

  //mdmod = modi + GeoPar.midray + 1;
  //mdg = g + GeoPar.midray + 1;
  //mdg1 = g1 + GeoPar.midray + 1;
  mdmod = modi + GeoPar.midray;        //(JD 11/06)
  mdg = g + GeoPar.midray;        //(JD 11/06)
  mdg1 = g1 + GeoPar.midray;        //(JD 11/06)

  anginy = SQR(GeoPar.pixsiz / GeoPar.radius);
  angiin = (REAL) GeoPar.midpix * anginy;
  sbtrin = (REAL) GeoPar.midpix * angiin;
  factor = (REAL) 0.25 * alfa * GeoPar.radius / Consts.pisq;
  //lpj = ((GeoPar.prjnum-1) / missp) * missp + 1;
  lpj = ((GeoPar.prjnum-1) / missp) * missp;    //new indexing (from zero) (JD)

  Anglst.getang(lpj, &beta0, &sinth, &costh);
  //Anglst.getang(1, &theta, &sinth, &costh);
  Anglst.getang(0, &theta, &sinth, &costh);    //new indexing (from zero) (JD)
    
  if((theta + Consts.twopi) <=  beta0) {

    fprintf(output, "\n         ***** illegal arrangement of projection angles *****");

    
 
  if(detecn != NULL) delete[] detecn ;
  if(q1 != NULL) delete[] q1 ;
  if(q2 != NULL) delete[] q2 ;
  if(ncos != NULL) delete[] ncos ;
  if(g != NULL) delete[] g ;
  if(g1 != NULL) delete[] g1;
  if(modi != NULL) delete[] modi ;           //wei 3/2005
  
    return TRUE;
  }

  // THIS IS WHERE THE RECONSTRUCTION STARTS
  // BUILD UP RECONSTRUCTED PICTURE ONE PROJECTION AT A TIME

  for(np = 0; np < GeoPar.prjnum; np += missp) {
    ProjFile.ReadProj(np, g, GeoPar.nrays);

    // CONVERT THE PROJECTION DATA TO ARC GEOMETRY BY LINEAR
    // INTERPOLATION IF NECESSARY

    if(!(GeoPar.arc || GeoPar.nrays <= 3)) {

      *mdg1 = *mdg;

      //for(i = 0; i < mraym2; i++){
      for(i = 1; i <= mraym2; i++){        //(JD 11/06)
        dtc = detecn[i-1];
        *(mdg1+i) = qintp(fmdray + dtc, g, GeoPar.nrays, 2);
        *(mdg1-i) = qintp(fmdray - dtc, g, GeoPar.nrays, 2);
      }

      for(i = 1; i < nraym; i++){
        g[i] = g1[i];
      }
    
    }

    // PRE-SMOOTHING OF PROJECTION DATA

    if(wconv != 1.0) {
      datm = g[0];
      dath = g[1];
      g[0] = wconv * datm + wconw * dath;

      //for(i = 1; i < nraym; i++) {
      for(i = 1; i < nraym; i++) {         //(JD 11/06)
        datl = datm;
        datm = dath;
        dath = g[i+1];
        g[i] = wconv * datm + wconw * (dath + datl);
      }

      g[GeoPar.nrays-1] = wconv * dath + wconw * datm;
    }

    // WEIGHTED PROJECTION DATA

    *(mdg1) = *(mdg);

    //for(i = 0; i < mdraym; i++) {
    for(i = 1; i <= mdraym; i++) {        //(JD 11/06)
      //*(mdg1+i) = *(mdg+i) * ncos[i];
      //*(mdg1-i) = *(mdg-i) * ncos[i];
      *(mdg1+i) = *(mdg+i) * ncos[i-1];      //(JD 11/06)
      *(mdg1-i) = *(mdg-i) * ncos[i-1];      //(JD 11/06)
    }

    // CONVOLUTION

    second(&cbeg);
    mdmod[0] = *q1 * (*mdg1) + (*q2) * (*mdg);
    if(nconr != 0) {

      //for(i = 0; i < limmid; i++) {
      for(i = 1; i <= limmid; i++) {      //(JD 11/06)
        mdmod[0] += *(q1+i) * ( *(mdg1+i) + *(mdg1-i) )+ *(q2+i) * ( *(mdg+i) + *(mdg-i));
      }
    }

    //for(k = 0; k < mdraym; k++) {
    for(k = 1; k <= mdraym; k++) {                 //(JD 11/06)
      mr1 = mdg1+k;
      ml1 = mdg1-k;
      mr = mdg+k;
      ml = mdg-k;
      pr1 = *q1 * (*mr1);
      pr2 = *q2 * (*mr);
      pl1 = *q1 * (*ml1);
      pl2 = *q2 * (*ml);
      if(nconr != 0) {

        //for(i = 0; i < nconr; i++) {
        for(i = 1; i <= nconr; i++) {              //(JD 11/06)
          if((k+i) > mdraym) { 

            //for(i1 = i; i1 < nconr; i1++) {
            for(i1 = i; i1 <= nconr; i1++) {       //(JD 11/06)

              //if(!((mr1-i1) <= g1)) {
              if(!((mr1-i1) <= (g1-1))) { //i.e. if (mr1-i1) <g1
                mq1 = q1 + i1;
                mq2 = q2 + i1;
                pr1 += *mq1 * *(mr1-i1);
                pr2 += *mq2 * *(mr -i1);
                pl1 += *mq1 * *(ml1+i1);
                pl2 += *mq2 * *(ml +i1);
                //break;        //(JD 11/06)
              }
            }
            break;
          }
          mq1 = q1 + i;
          mq2 = q2 + i;
          pr1 += *mq1 * (*(mr1+i) + *(mr1-i));
          pr2 += *mq2 * (*(mr +i) + *(mr -i));
          pl1 += *mq1 * (*(ml1+i) + *(ml1-i));
          pl2 += *mq2 * (*(ml +i) + *(ml -i));
        }

      }

      //*(mdmod + k) = pr1 + pr2 * ncos[k];
      //*(mdmod - k) = pl1 + pl2 * ncos[k];
      *(mdmod + k) = pr1 + pr2 * ncos[k-1];   // (JD 11/07/03)
      *(mdmod - k) = pl1 + pl2 * ncos[k-1];   // (JD 11/07/03)

    }

#ifdef FFCOMPARE
    fprintf(output,"\n xalg4: modi");
    for(i = 0; i < GeoPar.nrays; i++) {
      if((i % 3) == 0) fprintf(output,"\n");
      fprintf(output," %36.30f", (double) modi[i]);
    }
#endif

    second(&cend);
    ctime += cend-cbeg;

    // COMPUTE THE TRAPEZOIDAL WIDTH FOR BETA INTEGRATION

    beta1 = theta;
    sbeta = -costh;
    cbeta = sinth;
    if(np < lpj) Anglst.getang(np + missp, &theta, &sinth, &costh);
    if((np < lpj) && (theta <= beta1)) {
      fprintf(output, "\n         ***** illegal arrangement of projection angles *****");

  
  if(detecn != NULL) delete[] detecn ;
  if(q1 != NULL) delete[] q1 ;
  if(q2 != NULL) delete[] q2 ;
  if(ncos != NULL) delete[] ncos ;
  if(g != NULL) delete[] g ;
  if(g1 != NULL) delete[] g1;
  if(modi != NULL) delete[] modi ;           //wei 3/2005
  
      return TRUE;
    }

    //if(np >= lpj) Anglst.getang(1,&theta, &sinth, &costh);
    if(np >= lpj) Anglst.getang(0,&theta, &sinth, &costh);         // (JD 11/07/03)
    w = (REAL) 0.5 * (REAL) fmod(Consts.twopi + theta - beta0, Consts.twopi);

    beta0 = beta1;
    b = w * factor;

    // BACK PROJECTION

    cbhl = halfl * cbeta;
    sbhl = halfl * sbeta;
    cbpxs = GeoPar.pixsiz * cbeta;
    sbpxs = GeoPar.pixsiz * sbeta;
    usini = sbhl - cbhl;
    ucosi = GeoPar.radius - sbhl - cbhl;

    if(method != 0) {
      bovrc = b / rcube;
      wbkpjy = (GeoPar.radius + (REAL) 2.0 * (cbhl + sbhl)) * bovrc;
      wincy = (REAL) -2.0 * cbpxs * bovrc;
      wincx = (REAL) -2.0 * sbpxs * bovrc;

      if(method <= 0) {
        angsin = sbhl;
        angcos = GeoPar.radius - cbhl;
        anginc = cbpxs / GeoPar.radius - angiin;
        subtr = cbhl / GeoPar.radius - sbtrin;
      }
    }

    ind = 0;
    for(iy = 0; iy < GeoPar.nelem; iy++) {
      if(method >= 0) {
        usini -= sbpxs;
        usin = usini;
        ucosi += cbpxs;
        ucos = ucosi;
      }
      if(method != 0) {
        wbkpjy = wbkpjy + wincy;
        wbkpj = wbkpjy;
        if(method <= 0) {
          angsin -= sbpxs;
          angcos += cbpxs;
          anginc += anginy;
          subtr += angiin;
          ang = (REAL) atan2(angsin, angcos) - subtr;
        }
      }
      for(ix = 0; ix < GeoPar.nelem; ix++) {

        if(method >= 0) {

          usin += cbpxs;
          ucos += sbpxs;
          ang = (REAL) atan2(usin, ucos);
        }

        if(method == 0) {

          wbkpj = b / (usin * usin + ucos * ucos);
        }
        else {
          wbkpj += wincx;
          if(method < 0) ang += anginc;
        }
        kpr = fmdray - ang / alfa;
        recon[ind] += qintp(kpr, modi, GeoPar.nrays, interp) * wbkpj;
	ind++;
      }

    }

    second(&bend);
    btime += bend - cend;
  }
  fprintf(output, "\n          total time for obtaining convoluting function was %10.3f seconds", ftime); // changed precision to three digits - swr 1/21/06
  fprintf(output, "\n          total time for convolution was                    %10.3f seconds", ctime); // ditto
  fprintf(output, "\n          total time for back-projection was                %10.3f seconds", btime); // ditto

  if(detecn != NULL) delete[] detecn ;
  if(q1 != NULL) delete[] q1 ;
  if(q2 != NULL) delete[] q2 ;
  if(ncos != NULL) delete[] ncos ;
  if(g != NULL) delete[] g ;
  if(g1 != NULL) delete[] g1;
  if(modi != NULL) delete[] modi ;           //wei 3/2005
  
  return ret_xalg4;
}
     
