/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/exalg.cpp $
 $LastChangedRevision: 164 $
 $Date: 2015-05-14 00:49:54 -0400 (Thu, 14 May 2015) $
 $Author: zye $
 ***********************************************************

 EXALG DETERMINES WHICH ALGORITHM IS TO BE EXECUTED, WHICH OF
 THE STANDARD OPTIONS ARE IN EFFECT AND REPEATEDLY EXECUTES THE
 ALGORITHM UNTIL THE TERMINATION CRITERION (SET BY STOP COMMAND)
 IS SATISFIED.
 */

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include "blkdta.h"
#include "creacm.h"
#include "geom.h"
#include "term.h"
#include "raysel.h"
#include "uiod.h"
#include "int2str.h"
#include "second.h"
#include "bldlst.h"
#include "projfile.h"
#include "alb1.h"
#include "alb2.h"
#include "alb3.h"
#include "alb4.h"
#include "alb5.h"
#include "alp1.h"
#include "alp2.h"
#include "alp3.h"
#include "alp4.h"
#include "alp5.h"
#include "emap.h"
#include "lino.h"
#include "art.h"
#include "sart.h"
#include "conv.h"
#include "sirt.h"
#include "dcon.h"
#include "foru.h"
#include "rfl.h"
#include "back.h"
#include "quad.h"
#include "mart.h"
#include "trm1.h"
#include "trm2.h"
#include "trm3.h"
#include "trm4.h"
#include "trm5.h"
#include "smooth.h"
#include "contur.h"
#include "recfile.h"
#include "infile.h"
#include "blob.h"
#include "uiod.h"
#include "exalg.h"
#include "consts.h"
#include "bdhk.h"
#include "pseudo.h"
#include "superior.h"

using namespace std;

// macro to compute the number of elements in an array
#define length(array) (sizeof(array)/sizeof(array[0]))

static int totalIters = 0;

class NameTable
{
public:
	INTEGER Length;INTEGER* Names;

	INTEGER GetIndex(INTEGER word)
	{
		for (int i = 0; i < Length; i++)
		{
			if (word == Names[i])
			{
				return i;
			}
		}
		return -1;
	}

	INTEGER GetName(INTEGER index)
	{
		return Names[index];
	}
};

// Instantiate Algorithm classes
alb1_class alb1;
alb2_class alb2;
alb3_class alb3;
alb4_class alb4;
alb5_class alb5;
alp1_class alp1;
alp2_class alp2;
alp3_class alp3;
alp4_class alp4;
alp5_class alp5;
emap_class emap;
lino_class lino;
art_class art;
sart sart;
conv_class conv;
sirt_class sirt;
dcon_class dcon;
foru_class foru;
rfl_class rfl;
back_class back;
quad_class quad;
mart_class mart;
bdhk_class bdhk;

// Algorithms Classes
alg_class* AlgClasses[] =
{
	&alb1, // (ALB1) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alb2, // (ALB2) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alb3, // (ALB3) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alb4, // (ALB4) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alb5, // (ALB5) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alp1, // (ALP1) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alp2, // (ALP2) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alp3, // (ALP3) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alp4, // (ALP4) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&alp5, // (ALP5) DUMMY RECONSTRUCTION ALGORITHM - TO BE REPLACED BY USER
	&emap, // (EMAP) maximum a-posteriori probability expectation maximization algorithm
	&lino, // (LINO) LINOGRAM RECONSTRUCTION ALGORITHM
	&art,  // (ART) ALGEBRAIC RECONSTRUUCTION
	&sart, // (SART) SIMULTANEOUS ALGEBRAIC RECONSTRUUCTION TECHNIQUE
	&conv, // (CONV) GENERAL IMPLEMENTATION OF THE THREE BEST KNOWN CONVOLUTION METHODS
	&sirt, // (SIRT) IMPLEMENTATION OF THE SIRT ALGORITHM
	&dcon, // (DCON) A CONVOLUTION TYPE RECONSTRUCTION ALGORITHM FOR DIVERGENT
		   // BEAM PROJECTION DATA
	&foru, // (FOUR) FOURIER RECONSTRUCTION METHOD
	&rfl,  // (RFL) RHO-FILTERED LAYERGRAM RECONSTRUCTION METHOD
	&back, // (BACK) COMBINED BACK-PROJECTION RECONSTRUCTION ALGORITHM
	&quad, // (QUAD) QUADRATIC ITERATIVE RECONSTRUCTION ALGORITHMS
	&mart, // (MART) VARIOUS VERSIONS OF MULTIPLICATIVE ART ALGORITHM
	&bdhk  // (BDHK) ALGORITHM BASED ON BUTNARIU-DAVIDI-HERMAN-KAZANTSEV
		   // MINIMIZING TOTAL-VARIATION WITH PERTURBATIONS
};

#define NUMBER_OF_ALGORITHMS length(AlgClasses)

// Algorithms Names
static const INTEGER algn[NUMBER_OF_ALGORITHMS] =
{
	CHAR2INT('a', 'l', 'b', '1'),
	CHAR2INT('a', 'l', 'b', '2'),
	CHAR2INT('a', 'l', 'b', '3'),
	CHAR2INT('a', 'l', 'b', '4'),
	CHAR2INT('a', 'l', 'b', '5'),
	CHAR2INT('a', 'l', 'p', '1'),
	CHAR2INT('a', 'l', 'p', '2'),
	CHAR2INT('a', 'l', 'p', '3'),
	CHAR2INT('a', 'l', 'p',	'4'),
	CHAR2INT('a', 'l', 'p', '5'),
	CHAR2INT('e', 'm', 'a',	'p'),
	CHAR2INT('l', 'i', 'n', 'o'),
	CHAR2INT('a', 'r', 't',	' '),
	CHAR2INT('s', 'a', 'r', 't'),
	CHAR2INT('c', 'o', 'n', 'v'),
	CHAR2INT('s', 'i', 'r',	't'),
	CHAR2INT('d', 'c', 'o', 'n'),
	CHAR2INT('f', 'o', 'u',	'r'),
	CHAR2INT('r', 'f', 'l', ' '),
	CHAR2INT('b', 'a', 'c',	'k'),
	CHAR2INT('q', 'u', 'a', 'd'),
	CHAR2INT('m', 'a', 'r',	't'),
	CHAR2INT('b', 'd', 'h', 'k')
};

static NameTable Algorithms = { NUMBER_OF_ALGORITHMS, (int*) algn };

// Blob Algorithms Names
static const INTEGER algn_blob[] =
{
	CHAR2INT('a', 'l', 'b', '1'),
	CHAR2INT('a', 'l', 'b', '2'),
	CHAR2INT('a', 'l', 'b', '3'),
	CHAR2INT('a', 'l', 'b', '4'),
	CHAR2INT('a', 'l', 'b', '5'),
	CHAR2INT('e', 'm', 'a', 'p'),
	CHAR2INT('a', 'r', 't', ' '),
	CHAR2INT('s', 'a', 'r', 't'),
	CHAR2INT('s', 'i', 'r', 't'),
	CHAR2INT('q', 'u', 'a', 'd'),
	CHAR2INT('m', 'a', 'r', 't')
};

#define NUM_OF_ALG_BLOB length(algn_blob)

static NameTable BlobAlgorithms = { NUM_OF_ALG_BLOB, (int*) algn_blob };

void exalg()
{
	static const INTEGER exec_codes[30] =
	{
		//do not modify the first 4 entries!
		CHAR2INT('a', 'v', 'e', 'r'),
		CHAR2INT('z', 'e', 'r', 'o'),
		CHAR2INT('c', 'o', 'n', 't'),
		CHAR2INT('p', 'h', 'a', 'n'),
		//add any additional options to EXECUTE (first option) here

		CHAR2INT('a', 'l', 'b', '1'),
		CHAR2INT('a', 'l', 'b', '2'),
		CHAR2INT('a', 'l', 'b', '3'),
		CHAR2INT('a', 'l', 'b', '4'),
		CHAR2INT('a', 'l', 'b', '5'),
		CHAR2INT('a', 'l', 'p', '1'),
		CHAR2INT('a', 'l', 'p', '2'),
		CHAR2INT('a', 'l', 'p', '3'),
		CHAR2INT('a', 'l', 'p', '4'),
		CHAR2INT('a', 'l', 'p', '5'),
		CHAR2INT('e', 'm', 'a', 'p'),
		CHAR2INT('l', 'i', 'n', 'o'),
		CHAR2INT('a', 'r', 't', ' '),
		CHAR2INT('s', 'a', 'r', 't'),
		CHAR2INT('c', 'o', 'n', 'v'),
		CHAR2INT('s', 'i', 'r', 't'),
		CHAR2INT('d', 'c', 'o', 'n'),
		CHAR2INT('f', 'o', 'u', 'r'),
		CHAR2INT('r', 'f', 'l', ' '),
		CHAR2INT('b', 'a', 'c', 'k'),
		CHAR2INT('q', 'u', 'a', 'd'),
		CHAR2INT('m', 'a', 'r', 't'),
		CHAR2INT('b', 'd', 'h', 'k'),
		//add any additional reconstruction algorithms here

		CHAR2INT('s', 'm', 'o', 'o'),
		CHAR2INT('c', 'o', 'n', 't') };

	//these are variables indicating indexes into the exec_codes array above
	//the name ending in "St" indicates a starting index
	//the name ending in "Cnt" indicates number of entries in given category
	// !!! if you keep these indexes up to date after making changes to the
	// !!! array above, there is not need for changing indexes in the rest of
	// !!! the code below
	//there are four EXECUTE options starting at index 0 (these are the
	//first options - i.e. before the name of algorithm)
	INTEGER executeOp1St = 0;
	INTEGER executeOp1Cnt = 4;
	//there are 24 algorithm names starting at index 4
	INTEGER algSt = 4;
	INTEGER algCnt = 22;
	//there are 2 EXECUTE options starting at index 28 (these are the
	//second options - i.e. after the name of algorithm)
	INTEGER executeOp2St = 27;
	INTEGER executeOp2Cnt = 2;

	INTEGER word;
	BOOLEAN eol;
	BOOLEAN flagphan = FALSE;
	CHAR head[81];
	INTEGER count;

	INTEGER optsw;
	INTEGER optflg[51];

	INTEGER AlgIndex;

	BOOLEAN contin;

	INTEGER i;
	REAL val;
	INTEGER AlgName;
	REAL thresh;
	REAL w1;
	REAL w2;
	REAL w3;
	INTEGER iter;
	REAL begin, start;
	REAL end, end_alg;
	REAL time;

	BOOLEAN tflag;
	static BOOLEAN pix_old_flag = FALSE; // basis of the previous run
	static REAL delta_old; // delta of the previous blob run
	static REAL support_old; // support of the previous blob run
	static REAL shape_old; // shape of the previous blob run

	static BOOLEAN called = FALSE;
	static INTEGER* list;
	static REAL* weight;

	REAL* recon;
	REAL* recon_copy;
	INTEGER area; //size of picture
	INTEGER wx; //size of picture in x direct
	INTEGER wy; //size of picture in y direct

	if (ProjFile.Open("prjfil") != 0)
	{
		fprintf(output, "\n **** unable to open prjfil");
		fprintf(output, "\n **** program aborted\n");
		exit(-1);
	};

	ProjFile.ReadHeader(head, &NoisePar, &Spctrm, &Anglst);

	// PERFORM INITIALIZATION ONLY ON THE FIRST CALL
	if (!called)
	{

		Creacm.recon = new REAL[GeoPar.area];

		if (!Blob.pix_basis)
			Blob.recon = new REAL[Blob.area];

	};
	recon_copy = new REAL[GeoPar.area];

	RaySel.effcnt = 0;

	if (!RaySel.useray)
	{
		RaySel.numray = GeoPar.snrays;
	}
	else
	{
		RaySel.numray = GeoPar.usrays;
	}

	if (!RaySel.useray)
	{
		RaySel.initnr = GeoPar.fsnray;
		RaySel.endray = GeoPar.lsnray;
	}
	else
	{
		RaySel.initnr = GeoPar.fusray;
		RaySel.endray = GeoPar.lusray;
	}

	RaySel.begray = RaySel.initnr;
	RaySel.curnr = RaySel.initnr;

	if (RaySel.stype == 0)
	{
		RaySel.curnr = RaySel.initnr - RaySel.rayinc;
	}

	RaySel.initnp = 0;
	RaySel.curnp = RaySel.initnp;

	if (RaySel.stype == 1)
	{
		RaySel.curnp = RaySel.initnp - RaySel.prjinc;
	}

	RaySel.rcnt1 = 0;
	RaySel.rcnt2 = 0;

	// INITIALIZE OPTION FLAGS
	optsw = 0;
	for (i = 0; i < 50; i++)
	{
		optflg[i] = 0;
	}

	tflag = FALSE;

	word = InFile.getwrd(FALSE, &eol, exec_codes, executeOp1Cnt + algCnt);
	contin = FALSE;

	val = 0.0;

	if (word == exec_codes[0])
	{
		val = GeoPar.aveden;
		word = InFile.getwrd(FALSE, &eol, &(exec_codes[algSt]), algCnt);
	}
	else if (word == exec_codes[1])
	{
		word = InFile.getwrd(FALSE, &eol, &(exec_codes[algSt]), algCnt);
	}
	else if (word == exec_codes[2])
	{
		contin = TRUE;
		word = InFile.getwrd(FALSE, &eol, &(exec_codes[algSt]), algCnt);
	}
	else if (word == exec_codes[3])
	{
		flagphan = TRUE;
		if (Creacm.test == NULL)
		{
			fprintf(output,	"\n **** error: in conjuction with the option phantom, picture command's option must be test");
			fprintf(output, "\n **** algorithm execution aborted");
			return;
		}
		word = InFile.getwrd(FALSE, &eol, &(exec_codes[algSt]), algCnt);
	}

	// LOOK UP ROUTINE NAME
	// check algorithm name
	if ((AlgIndex = Algorithms.GetIndex(word)) < 0)
	{
		fprintf(output, "\n **** unknown algorithm %s", int2str(word));
		fprintf(output, "\n **** algorithm execution aborted");
		ProjFile.Close();
		return;
	}

	AlgName = word;

	BOOLEAN pix_basis = Blob.pix_basis;

	//check for algorithm that goes with blobs
	if (!Blob.pix_basis)
	{
		// check if selected algorithm is iterative
		if (BlobAlgorithms.GetIndex(AlgName) < 0)
		{
			fprintf(output,	"\n **** blob basis is not appropriate for algorithm %s, the algorithm is being executed using pixel basis instead.", int2str(AlgName));

			pix_basis = true;
			Blob.pix_basis = true;
		}
	}

	// ACCEPT AND DISPLAY COMMENT CARD
	InFile.GetComment(head);

	fprintf(output, "\n         %s", head);

	word = InFile.getwrd(FALSE, &eol, &(exec_codes[executeOp2St]),
			executeOp2Cnt);

	// CHECK TO SEE IF OPTION HAS BEEN SPECIFIED
	if (!eol)
	{
		if (word == exec_codes[executeOp2St])
		{
			//smooth
			optsw = 1;
		}

		if (word == exec_codes[executeOp2St + 1])
		{
			//contour
			optsw = 2;
		}

		thresh = InFile.getnum(TRUE, &eol);
		w1 = InFile.getnum(FALSE, &eol);
		w2 = InFile.getnum(FALSE, &eol);
		w3 = InFile.getnum(FALSE, &eol);

		// ECHO OPTION SELECTED
		if (optsw == 1)
		{
			fprintf(output, "\n         reconstruction is smoothed after");
		}

		if (optsw == 2)
		{
			fprintf(output, "\n         reconstruction is contoured after");
		}

		InFile.listit(optflg);

		if (optsw == 1)
		{
			if (pix_basis)
			{
				fprintf(output, "\n         threshold = %9.5f      w1 = %9.5f      w2 = %9.5f      w3 = %9.5f", thresh, w1, w2, w3);
			}
			else
			{
				fprintf(output,	"\n         threshold = %9.5f      w1 = %9.5f      w2 = %9.5f      w3 = %9.5f",	thresh, w1, w2, w3);

			}
		}

		if (optsw == 2)
		{
			fprintf(output,	"\n         threshold = %9.5f      w1 = %9.5f      w2 = %9.5f      w3 = %9.5f",	thresh, w1, w2, w3);
		}
	}

	if (contin)
	{
		if (!pix_old_flag && !pix_basis)
		{ // previously blob, now blob
			if (fabs(support_old - Blob.support) > Consts.zero || fabs(shape_old - Blob.shape) > Consts.zero || fabs(delta_old - Blob.delta) > Consts.zero)
			{
				fprintf(output,	"\n **** error: current blob parameters must match those of the previous blob basis execution");
				fprintf(output, "\n **** algorithm execution aborted");
				ProjFile.Close();
				return;
			}
		}

		if (!pix_old_flag && pix_basis)
		{ // previously blob, now pixel
			Blob.blob2pix(Blob.recon, Creacm.recon);
			delete[] Blob.recon;
		}

		if (pix_old_flag && !pix_basis)
		{ // previously pixel, now blob
			Blob.recon = new REAL[Blob.area];
			Blob.pix2blob(Creacm.recon, Blob.recon);
		}
	}
	else
	{
		if (!pix_basis)
			Blob.recon = new REAL[Blob.area];

		// initiate recon with val or phantom image
		if (flagphan == TRUE)
		{
			for (i = 0; i < GeoPar.area; i++)
			{
				Creacm.recon[i] = Creacm.test[i];
			}
			if (!pix_basis)
			{
				Blob.pix2blob(Creacm.recon, Blob.recon);
			}
		}
		else
		{
			if (pix_basis)
			{
				for (i = 0; i < GeoPar.area; i++)
				{
					Creacm.recon[i] = val;
				}
			}
			else
			{
				for (i = 0; i < Blob.area; i++)
				{
					Blob.recon[i] = val;
				}
			}
		}
	}

	if (pix_basis)
	{
		bldlst(&list, &weight);
		recon = Creacm.recon;
		wx = GeoPar.nelem;
		wy = GeoPar.nelem;
		area = GeoPar.area;
	}
	else
	{
		Blob.bbldlst(&list, &weight);
		recon = Blob.recon;
		wx = Blob.M;
		wy = Blob.H;
		area = Blob.area;
	}

	called = TRUE;

	second(&begin);
	start = begin;

	// initialize selected algorithm
	if (AlgClasses[AlgIndex]->Init() != 0)
	{
		fprintf(output, "\n **** algorithm %s failed to initialize", int2str(algn[AlgIndex]));
		fprintf(output, "\n **** algorithm execution aborted");
		ProjFile.Close();
		return;
	}

	count = 0;

	// run until the termination criterion is met or an error occurs
	while (TRUE)
	{
		count++;
		totalIters++;
		iter = ((count - 1) % 50) + 1;

		// execute single iteration of the selected algorithm
		if (SuperSet.superiorizationEnabled)
		{
			tflag = executeSuperiorization(recon, list, weight, count, AlgClasses[AlgIndex]);
		}
		else
		{
			tflag = AlgClasses[AlgIndex]->Run(recon, list, weight, count);
		}

		second(&end_alg);
		fprintf(output, "\n         algorithm executed in iteration %4i", count);
		time = end_alg - begin;
		fprintf(output,	"\n         %10.3f seconds for the execution of the algorithm",	time);

		// if error
		if (tflag)
		{
			fprintf(output, "\n **** algorithm indicated failure to continue");

			// reset algorithm after failure
			if (AlgClasses[AlgIndex]->Reset() != 0)
			{
				fprintf(output, "\n **** algorithm %s failed to reset",	int2str(algn[AlgIndex]));
			}

			fprintf(output, "\n **** algorithm execution aborted");
			break;
		}

		if (Term.termInstance == NULL)
		{
			tflag = ((count - Term.iterations) >= 0);
		}
		else
		{
			tflag = Term.termInstance->Run(recon, list, weight, count);
		}

		if (tflag)
		{
			// END THE ROUTINE IF THE TEST WAS SATISFIED
			// CHECK IF OPTION SPECIFIED FOR LAST ITERATION

			iter = 0;

			if (!pix_basis)
			{ // if currently blob, pixelize it

				Blob.recon = recon;
				Blob.blob2pix(Blob.recon, Creacm.recon);
			};

			if ((optsw == 1) && (optflg[iter] != 0))
			{
				smooth(Creacm.recon, GeoPar.nelem, GeoPar.nelem, thresh, w1, w2, w3);

#ifdef FFCOMPARE
				////////// dump recon ////////////
				fprintf(output, "\nexalg: algorithm %s result of iteration %i after smooth", int2str(algn[AlgIndex]), count);
				for (i = 0; i < area; i++)
				{
					if ((i % 3) == 0) fprintf(output, "\n");
					fprintf(output, " %36.30f", (double) recon[i]);
				}
				//////////////////////////////////
#endif
			};

			if ((optsw == 2) && (optflg[iter] != 0))
			{
				contur(Creacm.recon, GeoPar.nelem, GeoPar.nelem, thresh, w1, w2, w3);

#ifdef FFCOMPARE
				////////// dump recon ////////////
				fprintf(output, "\nexalg: algorithm %s result of iteration %i after contur", int2str(algn[AlgIndex]), count);
				for (i = 0; i < area; i++)
				{
					if ((i % 3) == 0) fprintf(output, "\n");
					fprintf(output, " %36.30f", (double) recon[i]);
				}
				//////////////////////////////////
#endif
			};

			RecFile.WriteRec(head, int2str(AlgName), count, iter, Creacm.recon);

			fprintf(output,	"\n         reconstruction completed after iteration %4i", count);

			// set value for the "previous basis" in the next execution : TRUE if pixel
			pix_old_flag = pix_basis;
			// if current execution is blob basis, record the values of the blob parameters,
			// so that IF the next iteration is also blob basis, they parameters should not be changed.
			if (!pix_basis)
			{
				support_old = Blob.support;
				shape_old = Blob.shape;
				delta_old = Blob.delta;
			}
			// reset algorithm after termination
			if (AlgClasses[AlgIndex]->Reset() != 0)
			{
				fprintf(output, "\n **** algorithm %s failed to reset",	int2str(algn[AlgIndex]));
			}

			break;
		}

		if (!pix_basis)
		{ // if currently blob, pixelize it
			Blob.recon = recon;
			Blob.blob2pix(Blob.recon, Creacm.recon);
		};

		for (i = 0; i < GeoPar.area; i++) recon_copy[i] = Creacm.recon[i];

		// EXECUTE SMOOTHING OR CONTOURING AS APPROPRIATE
		if ((optsw == 1) && (optflg[iter] != 0))
		{
			smooth(recon_copy, GeoPar.nelem, GeoPar.nelem, thresh, w1, w2, w3);

#ifdef FFCOMPARE
			////////// dump recon ////////////
			fprintf(output, "\nexalg: algorithm %s result of iteration %i after smooth", int2str(algn[AlgIndex]), count);
			for (i = 0; i < area; i++)
			{
				if ((i % 3) == 0) fprintf(output, "\n"); fprintf(output, " %36.30f", (double) recon[i]);
			}
			//////////////////////////////////
#endif
		};

		if ((optsw == 2) && (optflg[iter] != 0))
		{
			contur(recon_copy, GeoPar.nelem, GeoPar.nelem, thresh, w1, w2, w3);

			////////// dump recon ////////////
#ifdef FFCOMPARE
			fprintf(output, "\nexalg: algorithm %s result of iteration %i after contur", int2str(algn[AlgIndex]), count);
			for (i = 0; i < area; i++)
			{
				if ((i % 3) == 0) fprintf(output, "\n");
				fprintf(output, " %36.30f", (double) recon[i]);
			}
#endif
			//////////////////////////////////
		}

		RecFile.WriteRec(head, int2str(AlgName), count, iter, recon_copy);

		second(&end);
		fprintf(output, "\n         iteration %4i completed", count);
		time = end - begin;
		fprintf(output, "\n         %10.3f seconds for this iteration", time);
		begin = end;

	}

	second(&end);
	time = end - begin;
	fprintf(output, "\n         %10.3f seconds for this iteration", time);

	if (count > 1)
	{
		time = end - start;
		fprintf(output, "\n         %10.3f seconds for all iterations", time);
	};

	ProjFile.Close();

	if (list != NULL) delete[] list;
	if (weight != NULL)	delete[] weight;
	if (recon_copy != NULL)	delete[] recon_copy;

	return;
}
