/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/rtfort.cpp $
 $LastChangedRevision: 95 $
 $Date: 2014-07-02 19:43:36 -0400 (Wed, 02 Jul 2014) $
 $Author: agulati $
 ***********************************************************

 ************************************************************

 SUBROUTINE RTFORT(X, N, INVDIR)

 PURPOSE
 RTFORT COMPUTES THE HERMITIAN FOURIER SPECTRUM OF A ONE OR TWO
 DIMENSIONAL ARRAY OF REAL DATA (IF INVDIR=1), OR  THE REAL DATA
 CORRESPONDING TO THIS SPECTRUM (IF INVDIR=-1).

 USAGE
 TO COMPUTE DIRECT FOURIER TRANSFORM OF REAL ARRAY X (INVDIR=1)
 CALL RTFORT( X, N, INVDIR )
 HALF OF THE CONJUGATE-SYMMETRIC (HERMITIAN) SPECTRUM
 IS NOW STORED IN X AS COMPLEX FOURIER COEFFICIENTS.
 TO COMPUTE INVERSE TRANSFORM OF SEMI-SPECTRUM (INVDIR = -1)
 CALL RTFORT(X, N, INVDIR)
 ARRAY X NOW CONTAINS REAL DATA.

 PARAMETERS
 X
 ONE DIMENSIONAL, N(1)=1 : X CONTAINS 2*N(2) REAL NUMBERS IN
 CONSECUTIVE LOCATIONS AS INPUT TO DIRECT TRANSFORM OR AS RESULT
 OF INVERSE TRANSFORM.  THE SEMI-SPECTRUM CONSISTS OF ( N(2) + 1
 COMPLEX COEFFICIENTS STORED IN 2*N(2) + 2 LOCATIONS OF X.

 TWO DIMENSIONAL : X IS A REAL ARRAY OF (FAST*SLOW) DIMENSIONS
 2N(2) * N(1) AS INPUT TO DIRECT TRANSFORM OR AS RESULT OF
 INVERSE TRANSFORM.  THE SEMI-SPECTRUM CONSISTS OF
 ( N(2) + 1 ) * N(1)  (FAST*SLOW) COMPLEX FOURIER COEFFICIENTS
 STORED IN ( 2*N(2) + 2 ) * N(1) LOCATIONS OF X.

 N.B.  X MUST BE DIMENSIONED TO ACCOMODATE THE SEMI-SPECTRUM,
 WHICH REQUIRES MORE STORAGE THAN ITS CORRESPONDING REAL ARRAY.

 N
 A 3-ELEMENT VECTOR, N(1)= SLOW DIMENSION OF X ARRAY
 2*N(2) = FAST DIMENSION OF REAL X ARRAY (AN EVEN NUMBER)
 N(3) = 1 (ONLY 2-DIMENSIONAL ARRAYS ALLOWED)
 THE VALUES OF N(1) AND N(2) MUST BE COMPATIBLE WITH THE
 PARTICULAR COMPLEX FFT ROUTINE USED (E.G. SNFFT, HARM).

 INVDIR = 1   FOR DIRECT (FORWARD) TRANSFORM
 = -1  FOR INVERSE TRANSFORM

 SUBROUTINE CALLED:  TRANSM

 BASED ON R2FORT SUBROUTINE WRITTEN BY T.M.PETERS, JUNE 1971,
 FOR USE WITH RADIX TWO FFT.  RTFORT EXTENDS CAPABILITY TO
 RADIX ODD TRANSFORMS.        R.M. LEWITT, SEPT 1977.

 **************************************************************
 */

#include <cstdio>
#include <cmath>

#include "blkdta.h"
#include "snfft.h"
#include "transm.h"
#include "consts.h"

#include "rtfort.h"

void rtfort(REAL* x, INTEGER* nin, INTEGER invdir)
{

	INTEGER n[3];

	INTEGER m;
	INTEGER ns;
	INTEGER m1;
	INTEGER mm;
	INTEGER mm2;
	REAL sd;
	REAL cd;
	REAL sn;
	REAL cn;
	INTEGER if_1;
	INTEGER ib;
	REAL ct;
	REAL ca;
	INTEGER mtype;
	INTEGER index;
	INTEGER k;
	INTEGER ishift;
	INTEGER i;
	INTEGER igap;
	INTEGER jt;
	INTEGER i1;
	INTEGER j1;
	INTEGER j2;
	INTEGER k1;
	INTEGER nl;

	n[0] = nin[0];
	n[1] = nin[1];
	n[2] = 1;

	if (invdir == 1)
	{
		snfft(x, n, invdir);
	}

	m = n[1];
	ns = n[0];
	m1 = m + 1;
	mm = m + m;
	mm2 = mm + 2;

	// INITIALISE PARAMETERS FOR RECURSIVE GENERATION OF SINES
	// AND COSINES
	sd = Consts.pid2 / m;
	cd = ((REAL) 2.0) * SQR((REAL) sin(sd));

	sd = (REAL) -sin(sd + sd);

	// ASSUMING THE DIRECT FOURIER TRANSFORM HAS EXP(-I...)
	// IF DIRECT FT ROUTINE HAS EXP(+I...) FACTOR, THEN
	// REPLACE ABOVE STATEMENT BY
	// SD=SIN(SD+SD)

	sn = 0.;
	if (!(ns > 1))
	{

		// ONE DIMENSIONAL CASE
		if (!(invdir < 0))
		{

			// FORWARD TRANSFORM
			cn = 1.;

			// REPEAT FIRST TWO REAL NUMBERS AT ARRAY END FOR PERIODICITY.
			x[mm + 0] = x[0];
			x[mm + 1] = x[1];
		}
		else
		{

			// INVERSE TRANSFORM
			cn = -1.;
			sd = -sd;
		}

		// APPLY TRANSFORM SYMMETRY RELATIONS
		for (if_1 = 0; if_1 < m1; if_1 += 2)
		{
			ib = mm - if_1;
			transm(x, if_1, ib, sn, cn);
			ct = cn - (cd * cn + sd * sn);
			sn = sn + (sd * cn - cd * sn);
			ca = ((REAL) 0.5) / (SQR(ct) + SQR(sn)) + ((REAL) 0.5);
			sn = ca * sn;
			cn = ca * ct;
		}

		if (!(invdir > 0))
		{

			// FOR INVERSE TRANSFORM, ZERO LAST TWO ARRAY ELEMENTS
			x[mm + 0] = 0.;
			x[mm + 1] = 0.;
		}
	}
	else
	{
		// TWO DIMENSIONAL CASE
		// MTYPE = 0 (1)  FOR  M ODD (EVEN)
		mtype = (m / 2) * 2 - m + 1;
		if (!(invdir < 0))
		{
			cn = 1.;

			// FORWARD TRANSFORM. EXPAND ARRAY TO TRANSFORM FORMAT
			index = ns * mm;
			for (k = 1; k <= ns; k++)
			{
				ishift = (ns - k) * 2;
				for (i = 0; i < mm; i++)
				{
					index--;
					x[index + ishift] = x[index];
				}
				igap = index + ishift + mm;
				x[igap] = x[index];
				x[igap + 1] = x[index + 1];
			}
		}
		else
		{

			// INVERSE TRANSFORM
			cn = -1.;
			sd = -sd;
		}

		jt = (ns + 1) * mm2;
		for (i1 = 0; i1 < m; i1 += 2)
		{
			// K=0 CASE SEPARATELY
			j1 = i1;
			j2 = mm - i1;
			transm(x, j1, j2, sn, cn);

			index = 0;
			for (k1 = 1; k1 < ns; k1++)
			{
				index += mm2;
				j1 = index + i1;
				j2 = jt - j1 - 2;
				transm(x, j1, j2, sn, cn);
			}

			ct = cn - (cd * cn + sd * sn);
			sn = sn + (sd * cn - cd * sn);
			ca = ((REAL) 0.5) / (SQR(ct) + SQR(sn)) + ((REAL) 0.5);
			sn = ca * sn;
			cn = ca * ct;
		}

		if (!(mtype == 0))
		{

			// M EVEN, TAKE I1=M1 AS SPECIAL CASE
			m1 -= 1;             // decrement m1 - swr 11/25/04
			i1 = m1;
			transm(x, m1, m1, sn, cn);
			nl = ns / 2 + 1;
			index = 0;
			for (k1 = 2; k1 <= nl; k1++)
			{   // k => k1 - swr 11/25/04
				index += mm2;
				j1 = index + m1;
				j2 = jt - j1 - 2;  // decrement by 2 - swr 11/25/04
				transm(x, j1, j2, sn, cn);
			}
		}

		if (!(invdir > 0))
		{

			// INVERSE TRANSFORM, REVERT TO ORIGINAL FORMAT
			index = mm;
			for (k = 1; k < ns; k++)
			{
				ishift = k * 2;
				for (i = 1; i <= mm; i++)
				{
					//index++;
					x[index] = x[index + ishift];
					index++;
				}
			}

			// ZERO REMAINDER OF REAL ARRAY
			j1 = index;
			j2 = ns * mm2;
			for (i = j1; i < j2; i++)
			{
				x[i] = 0.;
			}
		}
	}

	if (invdir == -1)
	{
		snfft(x, n, invdir);
	}

	return;
}
