/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/signif1.c $
 $LastChangedRevision: 90 $
 $Date: 2014-07-02 17:31:12 -0400 (Wed, 02 Jul 2014) $
 $Author: olangthaler $
 ***********************************************************

 This function computes and stores the significance levels

 INPUTS:
 iexp - experiment number.
 nrun - number of runs in the current experiment.
 tot_PAIRS - number of phantoms with PAIRed structures used for a given experiment.
 compare - counter for the number of COMPAREs read in so far.
 niters - number of iterations to be compared.
 itr1 - array of length niters containing the iteration numbers for the first algorithm.
 itr2 - array of length niters containing the iteration numbers for the second algorithm.
 sigfl - file pointer to the file which stores the significance levels.
 fom - character string denoting the figure of merit.
 keywrd1 - string containing the keyword which defines the first algorithm.
 keywrd2 - string containing the keyword which defines the second algorithm.
 lowden - low clip value for pointwise accuracy calculation.
 highden - high clip value for pointwise accuracy calculation.
 */

#include <malloc.h>
#include <string.h>
#include <math.h>
#include <stdio.h>
#include "experimenter.h"
#include "read_fomfile.h"
#include "signif1.h"

void signif(int iexp, int nrun, int tot_PAIRS, int compare, int niters, int* itr1, int* itr2, FILE* sigfl, char* fom, char* keywrd1, char* keywrd2, float lowden, float highden)
{
	int i, j;
	double *fom1, *fom2, fom_diff, *sum_diff, *var, *sum_fom1, *sum_fom2;

	//  go here only if, for hit-ratio, there are no PAIRed structures
	if (strstr(fom, "Hit-ratio") != NULL)
	{
		if (tot_PAIRS == 0)
		{
			fprintf(sigfl, "\n \t \t FIGURE OF MERIT: %s\n", fom);
			fprintf(sigfl, "%s \t %s \t Significance Level\n", keywrd1,
					keywrd2);

			for (i = 0; i < niters; i++)
				fprintf(sigfl,
						"%d \t %d \t No paired structures in this experiment\n",
						itr1[i], itr2[i]);
			return;
		}
	}

	// allocate memory for various arrays

	// array to store figures of merit for algorithm 1
	fom1 = (double *) malloc(niters * sizeof(double));

	//  array to store figures of merit for algorithm 2
	fom2 = (double *) malloc(niters * sizeof(double));

	// array to store sum of the differences between the foms for the two algorithms
	sum_diff = (double *) malloc(niters * sizeof(double));

	// array to store sum of the squares of this difference
	var = (double *) malloc(niters * sizeof(double));

	// array to store sum of foms for algorithm 1
	sum_fom1 = (double *) malloc(niters * sizeof(double));

	// array to store sum of foms for algorithm 2
	sum_fom2 = (double *) malloc(niters * sizeof(double));

	for (j = 0; j < niters; j++) // initialize arrays
	{
		sum_diff[j] = 0.0;
		var[j] = 0.0;
		sum_fom1[j] = 0.0;
		sum_fom2[j] = 0.0;
	}

	for (i = 0; i < nrun; i++)
	{
		// read in foms and iteration numbers for both algorithms
		read_fomfile(iexp, i + 1, compare, niters, fom, fom1, fom2);
		for (j = 0; j < niters; j++)
		{
			/*
			 a) compute difference between foms of the two algorithms
			 b) sum these differences
			 c) sum the squares of these differences
			 d) sum the foms for algorithms 1 and 2
			*/

			fom_diff = fom1[j] - fom2[j];
			sum_diff[j] = sum_diff[j] + fom_diff;
			var[j] = var[j] + (fom_diff * fom_diff);
			sum_fom1[j] = sum_fom1[j] + fom1[j];
			sum_fom2[j] = sum_fom2[j] + fom2[j];
		}
	}

	// For hit-ratio, number of phantoms = tot_PAIRS. While for other foms number of phantoms = nrun

	if (strstr(fom, "Hit-ratio") != NULL)
	{
		nrun = tot_PAIRS;
	}

	for (j = 0; j < niters; j++)
	{
		// compute the average fom for algorithms 1 and 2
		sum_fom1[j] = sum_fom1[j] / nrun;
		sum_fom2[j] = sum_fom2[j] / nrun;

		if (var[j] == 0.0)
			sum_diff[j] = 0.5;
		else
		{
			// calculate significance levels. Note that if var[j]=0.0, the significance level for than iteration is 0.5.
			sum_diff[j] = sum_diff[j] / sqrt(var[j]);
			sum_diff[j] = 0.5 * erf(sum_diff[j] / sqrt(2.0));
			sum_diff[j] = 0.5 - dabs(sum_diff[j]);
		}
	}

	// print results and free memory
	if (strstr(fom, "Pointwise Accuracy") != NULL)
	{
		if (lowden == 0.0 && highden == 0.0) fprintf(sigfl, "\n \t FIGURE OF MERIT: %s.  NO CLIPPING\n", fom);
		else fprintf(sigfl, "\n \t FIGURE OF MERIT: %s.  CLIP VALUES ARE %.3f AND %.3f\n", fom, lowden, highden);
	}
	else fprintf(sigfl, "\n \t \t FIGURE OF MERIT: %s\n", fom);

	fprintf(sigfl, "%s \tmean\t\t\t%s\tmean\t\t\tsignificance level\n", keywrd1, keywrd2);
	for (i = 0; i < niters; i++) fprintf(sigfl, "%d\t%.20f  %d\t%.20f  %.20f\n", itr1[i], sum_fom1[i], itr2[i], sum_fom2[i], sum_diff[i]);

	free(fom1);
	free(fom2);
	free(sum_diff);
	free(var);
	free(sum_fom1);
	free(sum_fom2);
}

// function to return the absolute value
double dabs(double x)
{
	return x < 0 ? -x : x;
}
