/** @file displaylines.cpp
    @package snark14Display
    @author Bruno M. Carvalho and Deniz Sarioz
    licensed under (open-source) QPL v1.0
    which accompanies this distribution in the file QPL
        
    modification: Joanna Klukowska 1/13/2009
	    added choice of grayScale display for plot
	    in addition to color display
*/

#include "displaylines.hpp"
#include "verbosity.hpp"

#include <cstdio>
#include <cstdlib>
#include <iostream>

#include <qwidget.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qvariant.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qmessagebox.h>
#include <qfiledialog.h>
#include <qstring.h>
#include <qregexp.h>

/* 
 *  Constructs a displaylineswindow which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
 */
displaylineswindow::displaylineswindow( QWidget* parent,  const char* name, QPixmap* pixmap, bool modal, WFlags fl , int x, int y )
    : QDialog( parent, name, modal, fl )
{
  QPoint topleft;
  
  if ( !name ) 
    setName( "Column/Line Profile" );
  imagename=name;
  int outerXPAD = 10;
  int outerTopYPad = 10;
  int outerBotYPad = 90;
  int WID = pixmap->width();
  int WINWIDTH  = 2*outerXPAD + WID;
  int HEI = pixmap->height();
  int WINHEIGHT = HEI + outerTopYPad + outerBotYPad;
  resize( WINWIDTH, WINHEIGHT );
  setMinimumSize( QSize( WINWIDTH, WINHEIGHT ) );
  setMaximumSize( QSize( WINWIDTH, WINHEIGHT ) );
  //    setCaption( tr( "Column/Line Profile" ) );
  setCaption( name );
  topleft.setX(x);
  topleft.setY(y);
  this->move(topleft);
  localpixmap=pixmap;
  pixmapwidget = new QWidget( this, "pixmapwidget" );
  pixmapwidget->setGeometry( QRect( outerXPAD, outerTopYPad, WID, HEI ) );
  pixmapwidget->setBackgroundPixmap( *pixmap );
  int BUTTONWID = 100;
  int BUTTONHEI = 40;
  int BUTXPAD = 40;
  int BUTBOTPAD = 20;
  savebutton = new QPushButton( this, "savebutton" );
  savebutton->setGeometry( QRect( BUTXPAD, WINHEIGHT-BUTTONHEI-BUTBOTPAD, BUTTONWID, BUTTONHEI ) ); 
  savebutton->setCaption( tr( "" ) );
  savebutton->setText( tr( "Save" ) );
  QObject::connect(savebutton,SIGNAL(clicked()),this,SLOT(saveImage()));
  
  closebutton = new QPushButton( this, "closebutton" );
  closebutton->setGeometry( QRect( WINWIDTH-BUTTONWID-BUTXPAD, WINHEIGHT-BUTTONHEI-BUTBOTPAD, BUTTONWID, BUTTONHEI ) ); 
  closebutton->setCaption( tr( "" ) );
  closebutton->setText( tr( "Close" ) );
  QObject::connect(closebutton,SIGNAL(clicked()),this,SLOT(accept()));
} // displaylineswindow ctor

/*
 *  Destroys the object and frees any allocated resources
 */
displaylineswindow::~displaylineswindow()
{
  // no need to delete child widgets, Qt does it all for us
}

/** Saves the profile graph image
    @param void
    @author Bruno M. Carvalho, changes by Deniz
    @version 1.2 */ 
void displaylineswindow::saveImage()
{ 
  if(std::verbose >=2) std::cout << "++displaylineswindow::saveImage()" << std::endl;
  QString prefixForSaving = (imagename.simplifyWhiteSpace());
  if(std::verbose >=2) std::cout << "My prefix is originally: \"" << prefixForSaving << "\"" << std::endl;
  // beware weird characters in save file name!
  // these might be intentionally meaningful
  prefixForSaving.replace( QRegExp("\\="), "_EQ_");
  prefixForSaving.replace( QRegExp("\\+"), "_PLUS_");
  prefixForSaving.replace( QRegExp("\\@"), "_AT_");
  //  prefixForSaving.replace( QRegExp("-"), "_MINUS_"); // maybe
  // convert everything non-alphanumeric to underscore
  prefixForSaving.replace( QRegExp("[^a-zA-Z0-9]"), "_");
  // allow maximum of 2 consecutive underscores  
  prefixForSaving.replace( QRegExp("__+") , "__"); 
  // chuck underscore from beginning and end
  prefixForSaving.replace( QRegExp("^_+"), "");
  prefixForSaving.replace( QRegExp("_+$"), "");
  if(std::verbose >=2) std::cout << "My prefix became: \"" << prefixForSaving << "\"" << std::endl;
  // this is how the filter ought to be according to manual for version 3.3.
  QString myExtensionFilter =   // delimiter is double-semicolon, NOT '\n'
    "png (PNG file);;"
    "bmp (BMP file);;"
    "pbm (PBM file);;"
    "pgm (PGM file);;"
    "ppm (PPM file);;"
    "xbm (XBM file);;"
    "xpm (XPM file)"   ;
  QString mySaveAs = "Save Image as";
  QString startInDir = "./";
  QFileDialog fd(startInDir, myExtensionFilter, this, mySaveAs, true);
  fd.setMode(QFileDialog::AnyFile);
  fd.setSelection(prefixForSaving);
  if (QDialog::Accepted != fd.exec()) {
    if(std::verbose >=2) std::cout << "--displaylineswindow::saveImage() in a hurry" << std::endl;
    return; // happens, say, when cancel is pressed.  just leave, no harm.
  }
  QString userFileSavePrefix = fd.selectedFile();
  QString userFileSaveExtension = fd.selectedFilter();
  if (userFileSavePrefix.isEmpty()) {
    if(std::verbose >=2) std::cout << "--displaylineswindow::saveImage() in a hurry" << std::endl;
    return; // happens, say, when cancel is pressed.  just leave, no harm.
  }
  if(std::verbose >=2) std::cout << "Got file name as: \"" << userFileSavePrefix << "\"" << std::endl;
  if(std::verbose >=2) std::cout << "Got selected ext: \"" << userFileSaveExtension << "\"" << std::endl;
  /* NB: userFileSavePrefix now contains directory info as well.
   * if user entered additional forward-slashes without
   * knowing that it signified directory, that's NOT our problem.
   * (the user MAY legitimately enter "a/myImage" as file name if there is a directory "a" 
   *   within the currently open directory in the save dialog window)
   * do part only after last forward-slash (i.e., what user entered)...
   * don't want to burn ourselves if path in fact had legit non-alphanumericounderscore chars
   */
  QString justFilePrefix, justPath;
  // funnylooking because written to work with Qt BOTH 2.3 and 3.3
  int posLastSlash = userFileSavePrefix.findRev('/'); // returns pos if found, -1 if not.
  if (-1 != posLastSlash) {
    justPath = userFileSavePrefix.left(posLastSlash+1); // +1 to include last '/'
    justFilePrefix = userFileSavePrefix;
    QRegExp fakeRegex("^" + justPath); // it just *looks* fake...
    justFilePrefix.replace( fakeRegex, "" ); // chuck away including last '/' for justFilePrefix
    if(std::verbose >=2) std::cout << "just after regex replace, justFilePrefix: \"" << justFilePrefix << "\"" << std::endl;
  } else {  // if for some reason no forwardslash, handle that gracefully too.
    justFilePrefix = userFileSavePrefix;
    justPath = "";
  }
  if(std::verbose >=2) std::cout << "justPath is: \"" << justPath << "\"" << std::endl;
  if(std::verbose >=2) std::cout << "justFilePrefix is first: \"" << justFilePrefix << "\"" << std::endl;
  // remove justFilePrefix from ONLY the end of 'userFileSavePrefix'.
  justFilePrefix.replace( QRegExp("="), "_EQ_");
  justFilePrefix.replace( QRegExp("+"), "_PLUS_");
  justFilePrefix.replace( QRegExp("[^a-zA-Z0-9]"), "_"); // no forward slash in regex
  // allow maximum of 2 consecutive underscores  
  justFilePrefix.replace( QRegExp("__+") , "__"); 
  // chuck underscore from beginning and end
  justFilePrefix.replace( QRegExp("^_+"), "");
  justFilePrefix.replace( QRegExp("_+$"), "");
  if(std::verbose >=2) std::cout << "justFilePrefix is later: \"" << justFilePrefix << "\"" << std::endl;
  // reset userFileSavePrefix by putting back together justPath and the refined justFilePrefix
  userFileSavePrefix = justPath + justFilePrefix;
  userFileSaveExtension.truncate(3);
  if(std::verbose >=2) std::cout << "My userFileSavePrefix is \"" << userFileSavePrefix << "\"" << std::endl;
  if(std::verbose >=2) std::cout << "My userFileSaveExtension is \"" << userFileSaveExtension << "\"" << std::endl;
  QString extensionUpper = userFileSaveExtension.upper();
  QString saveFullNameWithPath = userFileSavePrefix + "." + extensionUpper;
  if(std::verbose >=2) std::cout << "extensionUpper is now \"" << extensionUpper << "\"" << std::endl;
  if(std::verbose >=2) std::cout << "saveFullNameWithPath is now \"" << saveFullNameWithPath << "\"" << std::endl;
  QImage prof=localpixmap->convertToImage();
  QImageIO recio;
  recio.setImage(prof);
  recio.setFormat(extensionUpper);
  recio.setFileName(saveFullNameWithPath);
  if ( recio.write() ) {
    if(std::verbose>=1) {
      std::cout << "successfully saved: \"" << saveFullNameWithPath << "\"" << std::endl;
    }
  } else {
    std::cerr << "UNsuccessfully tried to save as: \"" << saveFullNameWithPath << "\"" << std::endl;
  }
  if(std::verbose >=2) std::cout << "--displaylineswindow::saveImage()" << std::endl;
} // --displaylineswindow::saveImage()
